import time
import os
import pandas as pd
import tempfile
from shrynk.utils import md5, get_model_data


from wrapt_timeout_decorator import timeout as timeout_fn


class BaseCompressor:
    bench_exceptions = ()
    compression_options = []

    def __init__(self, model_name, *args, **kwargs):
        self.model_name = model_name
        self.model_data = None

    def infer(self, obj):
        raise NotImplementedError

    def _save(self, obj, fname, *args, **kwargs):
        raise NotImplementedError

    def save(self, obj, fname_prefix, inferred_kwargs=None):
        if inferred_kwargs is None:
            inferred_kwargs = self.infer(obj)
        return self._save(obj, fname_prefix, **inferred_kwargs)

    def load(self, path):
        raise NotImplementedError

    def get_features(self, *args, **kwargs):
        raise NotImplementedError

    def is_valid(self, *args, **kwargs):
        raise NotImplementedError

    def benchmark(self, object_save, kwargs, timeout):
        t1 = time.time()
        path = "ZZZklkl"
        with tempfile.TemporaryDirectory() as fdir:
            try:
                if timeout:
                    path = timeout_fn(timeout)(self.save)(object_save, fdir + "/", kwargs)
                else:
                    path = self.save(object_save, fdir + "/", kwargs)
            except TimeoutError:
                return None, None, None
            except self.bench_exceptions:
                return None, None, None
            try:
                write_time = time.time() - t1
                size = os.path.getsize(path)
                t1 - time.time()
                _ = self.load(path)
                read_time = time.time() - t1
            except self.bench_exceptions:
                return None, None, None
        return size, write_time, read_time

    def run_benchmarks(self, data_generator, save=True, ignore_seen=True, timeout=300):
        from preconvert.output import json

        model_path = os.path.expanduser("~/shrynk_{}.jsonl".format(self.model_name))
        if self.model_data is None:
            self.model_data = get_model_data(self.model_name)
        feature_ids = set([x["feature_id"] for x in self.model_data])
        results = []
        index = []
        for num, df in enumerate(data_generator):
            if isinstance(df, str) and os.path.isfile(df):
                try:
                    df = self.load(df)
                except self.bench_exceptions:
                    continue
            if df is None:
                continue
            if isinstance(df, str):
                continue
            stat_computation_time = time.time()
            features = self.get_features(df)
            if features is None:
                continue
            feature_id = md5(features)
            if ignore_seen and feature_id in feature_ids:
                print("seen", feature_id)
                continue
            stat_computation_time = time.time() - stat_computation_time
            result = {
                "feature_id": feature_id,
                "features": features,
                "class": self.__class__.__name__,
                "stat_computation_time": stat_computation_time,
            }
            bench = []
            for kwargs in self.compression_options:
                size, write_time, read_time = self.benchmark(df, kwargs, timeout)
                if size is None:
                    # write_error(line)
                    print("error, skipping", kwargs)
                    continue
                print(kwargs, size, write_time, read_time)
                bench.append(
                    {
                        "kwargs": json.dumps(kwargs),
                        "size": size,
                        "write_time": write_time,
                        "read_time": read_time,
                    }
                )
            result["bench"] = bench
            if bench:
                self.model_data.append(result)
                if save:
                    with open(model_path, "a") as f:
                        f.write(json.dumps(result) + "\n")
                results.append(result)

            feature_ids.add(feature_id)
        ### run benchmarks should return a total overview or something, but now just from the last df
        # return pd.DataFrame(bench).set_index("kwargs")
        return results



import time
import os
import random
import warnings
import json
import pkgutil
import pandas as pd
import numpy as np
import dill
from collections import defaultdict
from sklearn.ensemble import RandomForestClassifier
from shrynk.utils import scalers, get_model_data


class Predictor:
    def __init__(self, model_name, clf=RandomForestClassifier, **clf_kwargs):
        if "n_estimators" in clf._get_param_names() and "n_estimators" not in clf_kwargs:
            clf_kwargs["n_estimators"] = 100
        self.clf = clf(**clf_kwargs)
        self.model_name = model_name
        self.model_data = None
        self.X_ = None
        self.y_ = None

    def upsample(self, bests):
        largest_n = bests["y"].astype(str).value_counts().max()
        bests = (
            bests.groupby("y")
            .apply(lambda x: x.sample(largest_n, replace=True))
            .reset_index(drop=True)
        )
        return bests

    def train_test_split(self, bests, train_ratio=0.5):
        X_train = []
        y_train = []
        X_test = []
        y_test = []
        for name, group in bests.groupby("group_id"):
            if random.random() < train_ratio:
                X_train.append(pd.io.json.json_normalize(group["features"]).set_index(group.index))
                y_train.append(group.y)
            else:
                X_test.append(pd.io.json.json_normalize(group["features"]).set_index(group.index))
                y_test.append(group.y)
        X_train = pd.concat(X_train)
        y_train = pd.concat(y_train)
        X_test = pd.concat(X_test)
        y_test = pd.concat(y_test)
        return X_train, y_train, X_test, y_test

    def crossval(self, data, bests, target, n=5, train_ratio=0.5):
        results = defaultdict(list)
        for i in range(n):
            print("crossval", i + 1, "/", n)
            X_train, y_train, X_test, y_test = self.train_test_split(bests, train_ratio)
            return X_train, y_train, X_test, y_test
            self.clf.fit(X_train.fillna(-100), y_train)
            preds = self.clf.predict(X_test.fillna(-100))
            class_accuracy = np.mean(preds == y_test)
            # # # # self.X_, self.y_ = X, y
            print("predicting")
            for g_id, p, t in zip(X_test.index, preds, y_test):
                a = self.arg_lookup[p]
                b = self.arg_lookup[t]
                g = data[data.group_id == g_id]
                try:
                    ga = g[g.kwargs == a]
                    gb = g[g.kwargs == b]
                    results["pred"].extend(np.array(ga[target]) / np.array(gb[target]))
                except ValueError:
                    continue
            strategies = y_train.unique()
            for strategy in strategies:
                print("strategy", strategy)
                preds = [strategy] * len(y_test)
                for g_id, p, t in zip(X_test.index, preds, y_test):
                    a = self.arg_lookup[p]
                    b = self.arg_lookup[t]
                    g = data[data.group_id == g_id]
                    ga = g[g.kwargs == a]
                    gb = g[g.kwargs == b]
                    if ga.empty:
                        continue
                    results[strategy].extend(np.array(ga[target]) / np.array(gb[target]))
        report = defaultdict(dict)
        pred_mean = np.mean(results["pred"])
        for strategy in strategies:
            if strategy == "pred":
                continue
            strat_mean = np.mean(results[strategy])
            savings = (strat_mean - pred_mean) / pred_mean * 100
            strategy_name = str(self.arg_lookup[strategy])
            report[strategy_name] = savings
        return class_accuracy, report

    def train_model(self, size, write, read, scaler="z", n_validations=None, balanced=True):
        targets = ["size", "write_time", "read_time"]
        if self.model_data is None:
            self.model_data = get_model_data(self.model_name)
        scale = scalers.get(scaler, scaler)
        size_write_read = np.array((size, write, read))
        features = []
        y = []
        feature_ids = []
        for x in self.model_data:
            vals = [[y[t] for t in targets] for y in x["bench"]]
            z = (scale(vals) * size_write_read).sum(axis=1).argmin()
            features.append(x["features"])
            y.append(x["bench"][z]["kwargs"])
            feature_ids.append(x["feature_id"])

        bests = pd.DataFrame(features)
        bests["y"] = y
        bests["feature_id"] = feature_ids

        bests = bests.sort_values("y")
        bests["train"] = bests.index % 2 == 0

        if balanced:
            bests = self.upsample(bests)
        if n_validations is not None:
            train = bests.query("train")
            test = bests.query("~train")
            self.clf.fit(train.drop(["y", "train", "feature_id"], axis=1).fillna(-100), train["y"])
            preds = self.clf.predict(test.drop(["y", "train", "feature_id"], axis=1).fillna(-100))
            print("accuracy", np.mean(preds == test.y), "now refitting...")
        self.clf.fit(bests.drop(["y", "train", "feature_id"], axis=1).fillna(-100), bests.y)
        # TO FIX CROSSVAL
        # if n_validations is not None:
        #     return self.crossval(data, bests, target, n_validations)
        try:
            weight = "{}_{}_{}".format(*size_write_read)
            clf = dill.loads(
                pkgutil.get_data("data", "shrynk/{}_.pkl".format(self.model_name.lower(), weight))
            )
            self.clf = clf
            print("loaded model pickle")
        except FileNotFoundError:
            self.clf.fit(bests.drop(["y", "train", "feature_id"], axis=1).fillna(-100), bests.y)
        return self.clf

    def predict(self, features):
        if isinstance(features, pd.DataFrame):
            features = self.get_features(features)
        if isinstance(features, dict):
            features = pd.DataFrame([features])
        warnings.filterwarnings(module='sklearn*', action='ignore', category=DeprecationWarning)
        return json.loads(self.clf.predict(features.fillna(-100))[0])

    infer = predict

    # def save_model(self):

    # def load_model(self):
    # def eval_model(self):



import os
from gzip import GzipFile, decompress
import pkgutil
import hashlib
from sklearn.preprocessing import scale, robust_scale, minmax_scale, maxabs_scale
import functools


def md5(features):
    from preconvert.output import json

    return hashlib.md5(json.dumps(features, sort_keys=True).encode()).hexdigest()


scalers = {
    "z": scale,
    "scale": scale,
    "minmax_scale": minmax_scale,
    "maxabs_scale": maxabs_scale,
    "robust_scale": robust_scale,
}


@functools.lru_cache()
def get_model_data(model_name):
    from preconvert.output import json

    """ Gets the model data """
    try:
        data = pkgutil.get_data("data", "shrynk/{}.jsonl.gzip".format(model_name.lower()))
        data = [
            json.loads(line) for line in decompress(data).decode("utf8").split("\n") if line.strip()
        ]
        print("from package")
    except FileNotFoundError:
        with open(os.path.expanduser("~/shrynk_{}.jsonl".format(model_name))) as f:
            data = [json.loads(x) for x in f.read().split("\n") if x]
    return data


def gzip(file_path):
    with open(file_path) as fin:
        with GzipFile(file_path + ".gzip", "w") as fout:
            fout.write(fin.read().encode("utf8"))


def gunzip(file_path):
    with GzipFile(file_path + ".gzip", "r") as fin:
        with open(file_path, "w") as fout:
            fout.write(fin.read())


def _package_data(file_path):
    file_path = os.path.expanduser(file_path)
    gzip(file_path)
    base = os.path.basename(file_path).replace("shrynk_", "")
    dest = "/home/pascal/egoroot/shrynk/data/shrynk/"
    os.rename(file_path + ".gzip", os.path.join(dest, base + ".gzip"))



import re
import os
import zipfile

import numpy as np
import pandas as pd
from pyarrow import ArrowTypeError


from shrynk.compressor import BaseCompressor
from shrynk.predictor import Predictor


def safelen(x):
    try:
        return len(x)
    except TypeError:
        return 0


# story:
# in the setup of the compressor we should use the basics options that don't require extras_require
# in the extras we can add the ones we want, which will then be imported
# - default model is always included in the package
# - for extras, it will download the data and it will train a model?
# for models, set the random_seed to 42

from pandas.io.common import _compression_to_extension

_csv_opts = [
    {"engine": "csv", "compression": x} for x in [None] + list(_compression_to_extension.keys())
]

# OPTIONAL: load pyarrow
try:
    import pyarrow
    from pyarrow import ArrowTypeError

    _pyarrow_exceptions = (ArrowTypeError,)
    _pyarrow_opts = [
        [
            {"engine": "pyarrow", "compression": y}
            for y in re.split("[', {}]+", x.split(": ")[1])
            if y
        ]
        for x in pyarrow.compress.__doc__.split("\n")
        if "upported types" in x
    ][0]
except ImportError:
    arrow_exceptions = ()
    _pyarrow = []

# OPTIONAL: load fastparquet
try:
    from fastparquet.compression import compressions

    # BROTLI IS BUGGED!
    _fastparquet_opts = [
        {"engine": "fastparquet", "compression": x} for x in compressions.keys() if x != "BROTLI"
    ]
except ImportError:
    _fastparquet_opts = []


class PandasCompressor(Predictor, BaseCompressor):
    bench_exceptions = (
        ValueError,
        pd.errors.ParserError,
        zipfile.BadZipFile,
        UnicodeDecodeError,
        OSError,
        pd.errors.EmptyDataError,
    ) + _pyarrow_exceptions

    compression_options = _fastparquet_opts + _pyarrow_opts + _csv_opts
    # [

    #     {"engine": "csv", "compression": None},
    #     {"engine": "csv", "compression": "gzip"},
    #     {"engine": "csv", "compression": "bz2"},
    #     {"engine": "csv", "compression": "xz"},
    #     {"engine": "csv", "compression": "zip"},
    #     # pyarrow # {‘NONE’, ‘SNAPPY’, ‘GZIP’,  ‘BROTLI’, ‘LZ4’, ‘ZSTD’}
    #     {"engine": "pyarrow", "compression": None},
    #     {"engine": "pyarrow", "compression": "snappy"},
    #     {"engine": "pyarrow", "compression": "gzip"},
    #     {"engine": "pyarrow", "compression": "brotli"},
    #     {"engine": "fastparquet", "compression": "GZIP"},
    #     {"engine": "fastparquet", "compression": "UNCOMPRESSED"},
    #     {"engine": "fastparquet", "compression": "BROTLI"},
    #     # {"engine": "fastparquet", "compression": "LZ4"},
    #     # C
    #     # {"engine": "fastparquet", "compression": "LZO"},
    #     # # # # # # ("fastparquet", "ZSTANDARD"),
    #     # fastparquet can do per column
    #     # pip install fastparquet[brotli]
    #     # pip install fastparquet[lz4]
    #     # pip install fastparquet[lzo]
    #     # pip install fastparquet[zstandard]
    #     # ("fastparquet", {str(x): "BROTLI" if x % 2 == 1 else "GZIP" for x in range(5)})
    # ]

    @classmethod
    def infer_from_path(cls, file_path):
        endings = file_path.split(".")[-2:]
        engine = None
        compression = None
        for x in cls.compression_options:
            if x["engine"] in endings:
                engine = x["engine"]
            if x["compression"] in endings:
                compression = x["compression"]
        if engine is None and "parquet" in endings:
            engine = "auto"
        return {"engine": engine, "compression": compression}

    def _save(self, df, file_path_prefix, engine=None, compression=None, **save_kwargs):
        if file_path_prefix.endswith(".csv") or file_path_prefix.endswith(".parquet"):
            raise ValueError("Only provide a base path, not an extension, the path wil be returned")
        path = "{}.shrynk.{}.{}".format(file_path_prefix, engine, compression)
        if engine is None or engine == "csv":
            df.to_csv(path, **save_kwargs)
        else:
            df.columns = [str(x) for x in df.columns]
            df.to_parquet(path, engine=engine, compression=compression, **save_kwargs)
        return path

    @classmethod
    def load(cls, file_path, inferred_kwargs=None, **load_kwargs):
        if inferred_kwargs is None:
            inferred_kwargs = cls.infer_from_path(file_path)
        if inferred_kwargs["engine"] is None or inferred_kwargs["engine"] == "csv":
            data = pd.read_csv(file_path, **load_kwargs)
        else:
            data = pd.read_parquet(file_path, engine=inferred_kwargs["engine"], **load_kwargs)
        return data

    def get_features(self, df):
        num_cols = df.shape[1]
        num_obs = df.shape[0]
        float_cols = [c for c, d in zip(df.columns, df.dtypes) if "float" in str(d)]
        str_cols = [c for c, d in zip(df.columns, df.dtypes) if "object" in str(d)]
        if str_cols and not df[str_cols].empty:
            str_len_quantiles = list(
                df[str_cols].applymap(safelen).quantile([0.25, 0.50, 0.75], axis=0).mean(axis=1)
            )
            str_missing_proportion = (~df[str_cols].notnull()).mean().mean()
        else:
            str_len_quantiles = [0, 0, 0]
            str_missing_proportion = 0
        if float_cols and not df[float_cols].empty:
            float_equal_0 = (df[float_cols] == 0).mean().mean()
            float_missing_proportion = (~df[str_cols].notnull()).mean().mean()
        else:
            float_equal_0 = 0
            float_missing_proportion = 0
        cardinality = df.apply(pd.Series.nunique)
        cardinality_quantile_proportion = cardinality.quantile([0.25, 0.50, 0.75]) / num_obs
        memory_usage = df.memory_usage().sum()
        data = {
            # standard
            "num_obs": num_obs,
            "num_cols": num_cols,
            "num_float_vars": len(float_cols),
            "num_str_vars": len(str_cols),
            "percent_float": len(float_cols) / num_cols,
            "percent_str": len(str_cols) / num_cols,
            "str_missing_proportion": str_missing_proportion,
            "float_missing_proportion": float_missing_proportion,
            "cardinality_quantile_proportion_25": cardinality_quantile_proportion.iloc[0],
            "cardinality_quantile_proportion_50": cardinality_quantile_proportion.iloc[1],
            "cardinality_quantile_proportion_75": cardinality_quantile_proportion.iloc[2],
            # extra cpu
            "float_equal_0_proportion": float_equal_0,
            "str_len_quantile_25": str_len_quantiles[0],
            "str_len_quantile_50": str_len_quantiles[1],
            "str_len_quantile_75": str_len_quantiles[2],
            "memory_usage": memory_usage,
        }
        return data

    def is_valid(self, df):
        if df.shape[1] < 2:
            return None
        if df.empty:
            return None
        return df


# - set of ids
# X = [x["features"] for x in results]
# allowed_kwargs = set(['{"compression": null, "engine": "csv"}'])
# y = [min([y for y in x["bench"] if y["kwargs"] in allowed_kwargs], key=lambda x: x[target])["kwargs"] for x in results]
# -- for quick bench lookup
#

trained_models = {}


def _get_model(model_name, size_write_read):
    key = (model_name,) + size_write_read
    if key not in trained_models:
        pdc = PandasCompressor(model_name)
        pdc.train_model(*size_write_read)
        trained_models[key] = pdc
    return trained_models[key]


def save(
    df,
    fname_prefix,
    inferred_kwargs=None,
    size=3,
    write=1,
    read=1,
    model_name="default",
    **save_kwargs
):
    pdc = _get_model(model_name, (size, write, read))
    return pdc.save(df, fname_prefix, inferred_kwargs, **save_kwargs)


def load(file_path, inferred_kwargs=None, **load_kwargs):
    return PandasCompressor.load(file_path, inferred_kwargs, **load_kwargs)


def infer(df, size=1, write=0, read=0, model_name="default"):
    pdc = _get_model(model_name, (size, write, read))
    return pdc.infer(df)



from setuptools import find_packages
from setuptools import setup

with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

MAJOR_VERSION = '0'
MINOR_VERSION = '0'
MICRO_VERSION = '6'
VERSION = "{}.{}.{}".format(MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION)

setup(
    name='shrynk',
    version=VERSION,
    description="Using Machine Learning to learn how to Compress",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author='Pascal van Kooten',
    url='https://github.com/kootenpv/shrynk',
    author_email='kootenpv@gmail.com',
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        # '': ['*.txt', '*.rst'],
        'data': ['shrynk/*.jsonl', 'shrynk/*.pkl']
    },
    install_requires=[
        'pandas',
        'pyarrow',
        'fastparquet',
        "sklearn",
        "preconvert",
        "preconvert_numpy",
        'dill',
        'wrapt_timeout_decorator',
    ],
    extras_require={
        "all": ["fastparquet[brotli,lz4,lzo,snappy,zstandard]"],
        "noc": ["fastparquet[brotli,lz4,snappy]"],
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Customer Service',
        'Intended Audience :: System Administrators',
        'Operating System :: Microsoft',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Software Distribution',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
    license='MIT',
    packages=find_packages(),
    zip_safe=False,
    platforms='any',
)
